import numpy as np
import highspy

h = highspy.Highs()
h.setOptionValue("output_flag", False);
h.setOptionValue("presolve", "off");
inf = highspy.kHighsInf
lp = highspy.HighsLp()
lp.sense_ = highspy.ObjSense.kMaximize
lp.num_col_ = 5
lp.num_row_ = 1
lp.col_cost_ = np.array([8, 5, 3, 11, 7], dtype=np.double)
lp.col_lower_ = np.array([0, 0, 0, 0, 0], dtype=np.double)
lp.col_upper_ = np.array([1, 1, 1, 1, 1], dtype=np.double)
lp.row_lower_ = np.array([-inf], dtype=np.double)
lp.row_upper_ = np.array([11], dtype=np.double)
lp.a_matrix_.format_ = highspy.MatrixFormat.kRowwise
lp.a_matrix_.start_ = np.array([0, 5])
lp.a_matrix_.index_ = np.array([0, 1, 2, 3, 4])
lp.a_matrix_.value_ = np.array([4, 3, 1, 5, 4], dtype=np.double)
lp.integrality_ = np.array([highspy.HighsVarType.kInteger, highspy.HighsVarType.kInteger, highspy.HighsVarType.kInteger, highspy.HighsVarType.kInteger, highspy.HighsVarType.kInteger])
h.passModel(lp)

h.run()
solution = h.getSolution()
print(f"Solution: ({solution.col_value[0]}, {solution.col_value[1]}, {solution.col_value[2]}, {solution.col_value[3]}, {solution.col_value[4]})")

# Solution is [1, 0, 1, 1, 0]

# Illustrate setSolution

# First by passing back the optimal solution

h.clearSolver()
h.setSolution(solution)
h.run()
solution = h.getSolution()
print(f"Solution: ({solution.col_value[0]}, {solution.col_value[1]}, {solution.col_value[2]}, {solution.col_value[3]}, {solution.col_value[4]})")

# Now passing back the optimal values of two variables as a sparse solution
h.clearSolver()
index = np.array([0, 3])
value = np.array([1, 1], dtype=np.double)
h.setSolution(2, index, value)
h.run()
solution = h.getSolution()
print(f"Solution: ({solution.col_value[0]}, {solution.col_value[1]}, {solution.col_value[2]}, {solution.col_value[3]}, {solution.col_value[4]})")

# Test passing back the optimal value of one variable, and a non-optimal value of another, as a sparse solution in untyped array
h.clearSolver()
h.setSolution(2, np.array([0, 4]), np.array([1, 1]))
h.run()
solution = h.getSolution()
print(f"Solution: ({solution.col_value[0]}, {solution.col_value[1]}, {solution.col_value[2]}, {solution.col_value[3]}, {solution.col_value[4]})")





# The path to the MPS file instance assumes that this is run 
# from the root directory.

import highspy
import numpy as np

# Highs h
h = highspy.Highs()

# Solve from mps file
# Initialize an instance of Highs
# h = highspy.Highs()
# Here we are re-using the one from above.
h.readModel('check/instances/25fv47.mps')

# Print
lp = h.getLp()
num_nz = h.getNumNz()
print('LP has ', lp.num_col_, ' columns',
      lp.num_row_, ' rows and ', num_nz, ' nonzeros')

# Solve the problem
h.run()

# Print solution information
solution = h.getSolution()
basis = h.getBasis()
info = h.getInfo()
model_status = h.getModelStatus()
print('Model status = ', h.modelStatusToString(model_status))
print()
print('Optimal objective = ', info.objective_function_value)
print('Iteration count = ', info.simplex_iteration_count)
print('Primal solution status = ',
      h.solutionStatusToString(info.primal_solution_status))
print('Dual solution status = ',
      h.solutionStatusToString(info.dual_solution_status))
print('Basis validity = ', h.basisValidityToString(info.basis_validity))



from __future__ import annotations
from highspy._core import (
    BasisValidity,
    HessianFormat,
    HighsBasis,
    HighsBasisStatus,
    HighsHessian,
    HighsInfo,
    HighsInfoType,
    HighsLogType,
    HighsLp,
    HighsModel,
    HighsModelStatus,
    HighsObjectiveSolution,
    HighsOptionType,
    HighsOptions,
    HighsPresolveStatus,
    HighsRanging,
    HighsRangingRecord,
    HighsSolution,
    HighsSparseMatrix,
    HighsStatus,
    HighsVarType,
    MatrixFormat,
    ObjSense,
    SolutionStatus,
    cb,
    simplex_constants,
)
from . import _core
from .highs import Highs, HighsCallbackEvent, HighsCallback, HighspyArray, highs_var, highs_cons, highs_linear_expression

__all__: list[str] = [
    "__doc__",
    "ObjSense",
    "MatrixFormat",
    "HessianFormat",
    "SolutionStatus",
    "BasisValidity",
    "HighsModelStatus",
    "HighsPresolveStatus",
    "HighsBasisStatus",
    "HighsVarType",
    "HighsOptionType",
    "HighsInfoType",
    "HighsStatus",
    "HighsLogType",
    "HighsSparseMatrix",
    "HighsLp",
    "HighsHessian",
    "HighsModel",
    "HighsInfo",
    "HighsOptions",
    "Highs",
    "HighsSolution",
    "HighsObjectiveSolution",
    "HighsBasis",
    "HighsRangingRecord",
    "HighsRanging",
    "kHighsInf",
    "kHighsIInf",
    "HIGHS_VERSION_MAJOR",
    "HIGHS_VERSION_MINOR",
    "HIGHS_VERSION_PATCH",
    "simplex_constants",
    "cb",
    "kSolutionStatusNone",
    "kSolutionStatusInfeasible",
    "kSolutionStatusFeasible",
    "kBasisValidityInvalid",
    "kBasisValidityValid",
    "HighsCallbackEvent",
    "HighsCallback",
    "HighspyArray",
    "highs_var",
    "highs_cons",
    "highs_linear_expression"
]

HIGHS_VERSION_MAJOR: int
HIGHS_VERSION_MINOR: int
HIGHS_VERSION_PATCH: int
kBasisValidityInvalid: _core.BasisValidity
kBasisValidityValid: _core.BasisValidity
kHighsIInf: int
kHighsInf: float
kSolutionStatusFeasible: _core.SolutionStatus
kSolutionStatusInfeasible: _core.SolutionStatus
kSolutionStatusNone: _core.SolutionStatus



# The paths to MPS file instances assumes that this is run in the
# root directory of HiGHS

import numpy as np
import highspy

hscb = highspy.cb 

h = highspy.Highs()

# h.setOptionValue("log_to_console", True)

inf = highspy.kHighsInf
alt_inf = h.getInfinity()
print('highspy.kHighsInf = ', inf,
      'h.getInfinity() = ', alt_inf)

# ~~~
# Read in and solve avgas
h.readModel("check/instances/avgas.mps")

# h.writeModel("ml.mps")

h.run()
lp = h.getLp()
num_nz = h.getNumNz()
print("LP has ", lp.num_col_,
      " columns", lp.num_row_,
      " rows and ", num_nz, " nonzeros.")

# ~~~
# Clear so that incumbent model is empty
h.clear()

# Now define the blending model as a HighsLp instance
lp = highspy.HighsLp()
lp.num_col_ = 2
lp.num_row_ = 2
lp.sense_ = highspy.ObjSense.kMaximize
lp.col_cost_ = np.array([8, 10], dtype=np.double)
lp.col_lower_ = np.array([0, 0], dtype=np.double)
lp.col_upper_ = np.array([inf, inf], dtype=np.double)
lp.row_lower_ = np.array([-inf, -inf], dtype=np.double)
lp.row_upper_ = np.array([120, 210], dtype=np.double)
lp.a_matrix_.start_ = np.array([0, 2, 4])
lp.a_matrix_.index_ = np.array([0, 1, 0, 1])
lp.a_matrix_.value_ = np.array([0.3, 0.7, 0.5, 0.5], dtype=np.double)
h.passModel(lp)

# Solve
h.run()

# Print solution
solution = h.getSolution()
basis = h.getBasis()
info = h.getInfo()
model_status = h.getModelStatus()
print("Model status = ", h.modelStatusToString(model_status))
print("Optimal objective = ", info.objective_function_value)
print("Iteration count = ", info.simplex_iteration_count)
print(
    "Primal solution status = ", h.solutionStatusToString(
        info.primal_solution_status)
)
print("Dual solution status = ",
      h.solutionStatusToString(info.dual_solution_status))
print("Basis validity = ", h.basisValidityToString(info.basis_validity))
num_var = h.getNumCol()
num_row = h.getNumRow()
print("Variables")
for icol in range(num_var):
    print(icol, solution.col_value[icol],
          h.basisStatusToString(basis.col_status[icol]))
print("Constraints")
for irow in range(num_row):
    print(irow, solution.row_value[irow],
          h.basisStatusToString(basis.row_status[irow]))

# ~~~
# Clear so that incumbent model is empty
h.clear()

# Now define the test-semi-definite0 model (from TestQpSolver.cpp) 
# as a HighsModel instance
model = highspy.HighsModel()
model.lp_.model_name_ = "semi-definite"
model.lp_.num_col_ = 3
model.lp_.num_row_ = 1
model.lp_.col_cost_ = np.array([1.0, 1.0, 2.0], dtype=np.double)
model.lp_.col_lower_ = np.array([0, 0, 0], dtype=np.double)
model.lp_.col_upper_ = np.array([inf, inf, inf], dtype=np.double)
model.lp_.row_lower_ = np.array([2], dtype=np.double)
model.lp_.row_upper_ = np.array([inf], dtype=np.double)
model.lp_.a_matrix_.format_ = highspy.MatrixFormat.kColwise
model.lp_.a_matrix_.start_ = np.array([0, 1, 2, 3])
model.lp_.a_matrix_.index_ = np.array([0, 0, 0])
model.lp_.a_matrix_.value_ = np.array([1.0, 1.0, 1.0], dtype=np.double)
model.hessian_.dim_ = model.lp_.num_col_
model.hessian_.start_ = np.array([0, 2, 2, 3])
model.hessian_.index_ = np.array([0, 2, 2])
model.hessian_.value_ = np.array([2.0, -1.0, 1.0], dtype=np.double)

print("test-semi-definite0 as HighsModel")
h.passModel(model)
h.run()

# ~~~
# Clear so that incumbent model is empty
h.clear()
num_col = 3
num_row = 1
sense = highspy.ObjSense.kMinimize
offset = 0
col_cost = np.array([1.0, 1.0, 2.0], dtype=np.double)
col_lower = np.array([0, 0, 0], dtype=np.double)
col_upper = np.array([inf, inf, inf], dtype=np.double)
row_lower = np.array([2], dtype=np.double)
row_upper = np.array([inf], dtype=np.double)
a_matrix_format = highspy.MatrixFormat.kColwise
a_matrix_start = np.array([0, 1, 2, 3])
a_matrix_index = np.array([0, 0, 0])
a_matrix_value = np.array([1.0, 1.0, 1.0], dtype=np.double)
a_matrix_num_nz = a_matrix_start[num_col]
hessian_format = highspy.HessianFormat.kTriangular
hessian_start = np.array([0, 2, 2, 3])
hessian_index = np.array([0, 2, 2])
hessian_value = np.array([2.0, -1.0, 1.0], dtype=np.double)
hessian_num_nz = hessian_start[num_col]
integrality = np.array([0, 0, 0])

print("test-semi-definite0 as pointers")
h.passModel(
    num_col,
    num_row,
    a_matrix_num_nz,
    hessian_num_nz,
    a_matrix_format,
    hessian_format,
    sense,
    offset,
    col_cost,
    col_lower,
    col_upper,
    row_lower,
    row_upper,
    a_matrix_start,
    a_matrix_index,
    a_matrix_value,
    hessian_start,
    hessian_index,
    hessian_value,
    integrality,
)
h.run()
h.writeSolution("", 1)

# ~~~
# Clear so that incumbent model is empty
h.clear()
print("25fv47 as HighsModel")

h.readModel("check/instances/25fv47.mps")

h.presolve()
presolved_lp = h.getPresolvedLp()

# Create a HiGHS instance to solve the presolved LP
print('\nCreate Highs instance to solve presolved LP')
h1 = highspy.Highs()
h1.passModel(presolved_lp)

# Get and set options
options = h1.getOptions()
options.presolve = "off"
options.solver = "ipm"

h1.passOptions(options)

# can be used to check option values
# h1.writeOptions("")

h1.run()
solution = h1.getSolution()
basis = h1.getBasis()

print("Crossover, then postsolve using solution and basis from another instance")

h.postsolve(solution, basis)

# Get solution
info = h.getInfo()
model_status = h.getModelStatus()
print("Model status = ", h.modelStatusToString(model_status))
print("Optimal objective = ", info.objective_function_value)
print("Iteration count = ", info.simplex_iteration_count)

run_time = h.getRunTime()
print("Total HiGHS run time is ", run_time)

# Get an optimal basis

# Clear so that incumbent model is empty
h.clear()
print("25fv47 as HighsModel")

h.readModel("check/instances/25fv47.mps")

h.run()
simplex_iteration_count = h.getInfo().simplex_iteration_count
print("From initial basis, simplex iteration count =", simplex_iteration_count)
basis = h.getBasis()
h.clearSolver()

h.setBasis(basis)
h.run()
simplex_iteration_count = h.getInfo().simplex_iteration_count
print("From optimal basis, simplex iteration count =", simplex_iteration_count)
status = h.setBasis()
h.run()
simplex_iteration_count = h.getInfo().simplex_iteration_count
print("From logical basis, simplex iteration count =", simplex_iteration_count)


# Define a callback

def user_interrupt_callback(
    callback_type,
    message,
    data_out,
    data_in,
    user_callback_data
):
    # dev_run = True
    dev_run = False

    # Constants for iteration limits or objective targets, adjust as required
    SIMPLEX_ITERATION_LIMIT = 100
    IPM_ITERATION_LIMIT = 100
    EGOUT_OBJECTIVE_TARGET = 1.0

    # Callback for MIP Improving Solution
    if callback_type == hscb.HighsCallbackType.kCallbackMipImprovingSolution:
        # Assuming it is a list or array
        assert user_callback_data is not None, "User callback data is None!"
        local_callback_data = user_callback_data[0]

        if dev_run:
            print(f"userCallback(type {callback_type};")
            print(f"data {local_callback_data:.4g}): {message}")
            print(f"with objective {data_out.objective_function_value}")
            print(f"and solution[0] = {data_out.mip_solution[0]}")

        # Check and update the objective function value
        assert (
            local_callback_data >= data_out.objective_function_value
        ), "Objective function value is invalid!"
        user_callback_data[0] = data_out.objective_function_value

    else:
        # Various other callback types
        if callback_type == hscb.HighsCallbackType.kCallbackLogging:
            if dev_run:
                print(f"userInterruptCallback(type {callback_type}): {message}")

        elif callback_type == hscb.HighsCallbackType.kCallbackSimplexInterrupt:
            if dev_run:
                print(f"userInterruptCallback(type {callback_type}): {message}")
                print("with iteration count = ", 
                      data_out.simplex_iteration_count)

            data_in.user_interrupt = (
                data_out.simplex_iteration_count > SIMPLEX_ITERATION_LIMIT
            )

        elif callback_type == hscb.HighsCallbackType.kCallbackIpmInterrupt:
            if dev_run:
                print(f"userInterruptCallback(type {callback_type}): {message}")
                print(f"with iteration count = {data_out.ipm_iteration_count}")

            data_in.user_interrupt = (
                data_out.ipm_iteration_count > IPM_ITERATION_LIMIT
            )

        elif callback_type == hscb.HighsCallbackType.kCallbackMipInterrupt:
            if dev_run:
                print(f"userInterruptCallback(type {callback_type}): {message}")
                print(f"Dual bound = {data_out.mip_dual_bound:.4g}")
                print(f"Primal bound = {data_out.mip_primal_bound:.4g}")
                print(f"Gap = {data_out.mip_gap:.4g}")
                print(f"Objective = {data_out.objective_function_value:.4g}")

            data_in.user_interrupt = (
                data_out.objective_function_value < EGOUT_OBJECTIVE_TARGET
            )


# Define model
h.addVar(-inf, inf)
h.addVar(-inf, inf)
h.changeColsCost(2, np.array([0, 1]), np.array([0, 1], dtype=np.double))
num_cons = 2
lower = np.array([2, 0], dtype=np.double)
upper = np.array([inf, inf], dtype=np.double)
num_new_nz = 4
starts = np.array([0, 2])
indices = np.array([0, 1, 0, 1])
values = np.array([-1, 1, 1, 1], dtype=np.double)
h.addRows(num_cons, lower, upper, num_new_nz, starts, indices, values)


# Set callback and run
h.setCallback(user_interrupt_callback, None)
h.startCallback(hscb.HighsCallbackType.kCallbackLogging)

h.run()
h.stopCallback(hscb.HighsCallbackType.kCallbackLogging)

# Get solution
num_var = h.getNumCol()
solution = h.getSolution()
basis = h.getBasis()
info = h.getInfo()
model_status = h.getModelStatus()
print("Model status = ", h.modelStatusToString(model_status))
print("Optimal objective = ", info.objective_function_value)
print("Iteration count = ", info.simplex_iteration_count)
print(
    "Primal solution status = ", h.solutionStatusToString(
        info.primal_solution_status)
)
print("Dual solution status = ",
      h.solutionStatusToString(info.dual_solution_status))
print("Basis validity = ", h.basisValidityToString(info.basis_validity))
print("Variables:")
for icol in range(0, 5):
    print(icol, solution.col_value[icol],
          h.basisStatusToString(basis.col_status[icol]))
print("...")
for icol in range(num_var-2, num_var):
    print(icol, solution.col_value[icol],
          h.basisStatusToString(basis.col_status[icol]))

print("computing IIS for lp-incompatible-bounds")
"""
LP has row0 and col2 with inconsistent bounds.

When prioritising rows, row0 and its constituent columns (1, 2) should be found
When prioritising columns, col2 and its constituent rows (0, 1) should be found
"""
# Define the LP
lp = highspy.HighsLp()
lp.num_col_ = 3
lp.num_row_ = 2
lp.col_cost_ = np.array([0, 0, 0], dtype=np.double)
lp.col_lower_ = np.array([0, 0, 0], dtype=np.double)
lp.col_upper_ = np.array([1, 1, -1], dtype=np.double)
lp.row_lower_ = np.array([1, 0], dtype=np.double)
lp.row_upper_ = np.array([0, 1], dtype=np.double)
lp.a_matrix_.format_ = highspy.MatrixFormat.kRowwise
lp.a_matrix_.start_ = np.array([0, 2, 4])
lp.a_matrix_.index_ = np.array([1, 2, 0, 2])
lp.a_matrix_.value_ = np.array([1, 1, 1, 1], dtype=np.double)

h.clear()
h.passModel(lp)
h.run()
assert h.getModelStatus() == highspy.HighsModelStatus.kInfeasible

# Set IIS strategy to row priority and get IIS
h.setOptionValue("iis_strategy", highspy.IisStrategy.kIisStrategyFromLpRowPriority)

iis = highspy.HighsIis()
assert h.getIis(iis) == highspy.HighsStatus.kOk
assert len(iis.col_index) == 0
assert len(iis.row_index) == 1
assert iis.row_index[0] == 0
assert iis.row_bound[0] == highspy.IisBoundStatus.kIisBoundStatusBoxed

# Set IIS strategy to column priority and get IIS
h.setOptionValue("iis_strategy", highspy.IisStrategy.kIisStrategyFromLpColPriority)
iis.invalidate()
assert h.getIis(iis) == highspy.HighsStatus.kOk
assert len(iis.col_index) == 1
assert len(iis.row_index) == 0
assert iis.col_index[0] == 2
assert iis.col_bound[0] == highspy.IisBoundStatus.kIisBoundStatusBoxed

print("IIS computation completed successfully")

print("computing feasibility relaxation")
h.clear()
inf = h.getInfinity()

num_col = 2
num_row = 3
num_nz = 6
a_format = highspy.MatrixFormat.kColwise
sense = highspy.ObjSense.kMinimize
offset = 0
col_cost = np.array([1, -2], dtype=np.double)
col_lower = np.array([5, -inf], dtype=np.double)
col_upper = np.array([inf, inf], dtype=np.double)
row_lower = np.array([2, -inf, -inf], dtype=np.double)
row_upper = np.array([inf, 1, 20], dtype=np.double)
a_start = np.array([0, 3])
a_index = np.array([0, 1, 2, 0, 1, 2])
a_value = np.array([-1, -3, 20, 21, 2, 1], dtype=np.double)
integrality = np.array([highspy.HighsVarType.kInteger, highspy.HighsVarType.kInteger])

h.passModel(
    num_col, num_row, num_nz, a_format, sense, offset,
    col_cost, col_lower, col_upper,
    row_lower, row_upper,
    a_start, a_index, a_value,
    integrality
)

assert h.feasibilityRelaxation(1, 1, 1) == highspy.HighsStatus.kOk

info = h.getInfo()
objective_function_value = info.objective_function_value

solution = h.getSolution()

assert abs(objective_function_value - 5) < 1e-6, f"Expected objective value 5, got {objective_function_value}"
assert abs(solution.col_value[0] - 1) < 1e-6, f"Expected solution[0] = 1, got {solution.col_value[0]}"
assert abs(solution.col_value[1] - 1) < 1e-6, f"Expected solution[1] = 1, got {solution.col_value[1]}"

print("Feasibility Relaxation Test Passed")

# Using infeasible LP from AMPL documentation
h = highspy.Highs()
lp = highspy.HighsLp()
lp.num_col_ = 2
lp.num_row_ = 3
lp.col_cost_ = np.array([1, -2], dtype=np.double)
lp.col_lower_ = np.array([5, -h.getInfinity()], dtype=np.double)
lp.col_upper_ = np.array([h.getInfinity(), h.getInfinity()], dtype=np.double)
lp.col_names_ = ["X", "Y"]
lp.row_lower_ = np.array([2, -h.getInfinity(), -h.getInfinity()], dtype=np.double)
lp.row_upper_ = np.array([h.getInfinity(), 1, 20], dtype=np.double)
lp.row_names_ = ["R0", "R1", "R2"]
lp.a_matrix_.start_ = np.array([0, 3, 6])
lp.a_matrix_.index_ = np.array([0, 1, 2, 0, 1, 2])
lp.a_matrix_.value_ = np.array([-1, -3, 20, 21, 2, 1], dtype=np.double)
lp.integrality_ = np.array([highspy.HighsVarType.kInteger, highspy.HighsVarType.kInteger])
h.setOptionValue("output_flag", False)
h.passModel(lp)

# Vanilla feasibility relaxation
print("Vanilla feasibility relaxation")
h.feasibilityRelaxation(1, 1, 1)
solution = h.getSolution()
print(f"Solution: ({solution.col_value[0]}, {solution.col_value[1]})")
print(f"Slacks: ({solution.row_value[0] - lp.row_lower_[0]}, "
      f"{lp.row_upper_[1] - solution.row_value[1]}, "
      f"{lp.row_upper_[2] - solution.row_value[2]})")

# Respect all lower bounds
print("\nRespect all lower bounds")
h.feasibilityRelaxation(-1, 1, 1)
solution = h.getSolution()
print(f"Solution: ({solution.col_value[0]}, {solution.col_value[1]})")
print(f"Slacks: ({solution.row_value[0] - lp.row_lower_[0]}, "
      f"{lp.row_upper_[1] - solution.row_value[1]}, "
      f"{lp.row_upper_[2] - solution.row_value[2]})")

# Local penalties RHS {1, -1, 10}
print("\nLocal penalties RHS {1, -1, 10}")
local_rhs_penalty = np.array([1, -1, 10], dtype=np.double)
h.feasibilityRelaxation(1, 1, 0, None, None, local_rhs_penalty)
solution = h.getSolution()
print(f"Solution: ({solution.col_value[0]}, {solution.col_value[1]})")
print(f"Slacks: ({solution.row_value[0] - lp.row_lower_[0]}, "
      f"{lp.row_upper_[1] - solution.row_value[1]}, "
      f"{lp.row_upper_[2] - solution.row_value[2]})")

# Local penalties RHS {10, 1, 1}
print("\nLocal penalties RHS {10, 1, 1}")
local_rhs_penalty = np.array([10, 1, 1], dtype=np.double)
h.feasibilityRelaxation(1, 1, 0, None, None, local_rhs_penalty)
solution = h.getSolution()
print(f"Solution: ({solution.col_value[0]}, {solution.col_value[1]})")
print(f"Slacks: ({solution.row_value[0] - lp.row_lower_[0]}, "
      f"{lp.row_upper_[1] - solution.row_value[1]}, "
      f"{lp.row_upper_[2] - solution.row_value[2]})")

iis = highspy.HighsIis()
assert h.getIis(iis) == highspy.HighsStatus.kOk

print("\nIIS")
print("row_index:", iis.row_index)
print("row_bound:", iis.row_bound)

print("col_index:", iis.col_index)
print("col_bound:", iis.col_bound)

# ~~~
# Clear so that incumbent model is empty
h.clear()



"""
Callback interface submodule
"""

from __future__ import annotations
import highspy._core
import typing

__all__ = [
    "HighsCallbackDataIn",
    "HighsCallbackDataOut",
    "HighsCallbackType",
    "kCallbackIpmInterrupt",
    "kCallbackLogging",
    "kCallbackMax",
    "kCallbackMin",
    "kCallbackMipDefineLazyConstraints",
    "kCallbackMipGetCutPool",
    "kCallbackMipImprovingSolution",
    "kCallbackMipInterrupt",
    "kCallbackMipLogging",
    "kCallbackMipSolution",
    "kCallbackSimplexInterrupt",
    "kNumCallbackType",
]

class HighsCallbackDataIn:
    user_interrupt: int
    def __init__(self) -> None: ...

class HighsCallbackDataOut:
    ipm_iteration_count: int
    log_type: int
    mip_dual_bound: float
    mip_gap: float
    mip_node_count: int
    mip_primal_bound: float
    objective_function_value: float
    pdlp_iteration_count: int
    running_time: float
    simplex_iteration_count: int
    def __init__(self) -> None: ...
    @property
    def mip_solution(self) -> highspy._core.readonly_ptr_wrapper_double: ...

class HighsCallbackType:
    """
    Members:
      kCallbackMin
      kCallbackLogging
      kCallbackSimplexInterrupt
      kCallbackIpmInterrupt
      kCallbackMipSolution
      kCallbackMipImprovingSolution
      kCallbackMipLogging
      kCallbackMipInterrupt
      kCallbackMipGetCutPool
      kCallbackMipDefineLazyConstraints
      kCallbackMax
      kNumCallbackType
    """

    __members__: typing.ClassVar[dict[str, HighsCallbackType]]
    kCallbackIpmInterrupt: typing.ClassVar[HighsCallbackType]
    kCallbackLogging: typing.ClassVar[HighsCallbackType]
    kCallbackMax: typing.ClassVar[HighsCallbackType]
    kCallbackMin: typing.ClassVar[HighsCallbackType]
    kCallbackMipDefineLazyConstraints: typing.ClassVar[HighsCallbackType]
    kCallbackMipGetCutPool: typing.ClassVar[HighsCallbackType]
    kCallbackMipImprovingSolution: typing.ClassVar[HighsCallbackType]
    kCallbackMipInterrupt: typing.ClassVar[HighsCallbackType]
    kCallbackMipLogging: typing.ClassVar[HighsCallbackType]
    kCallbackMipSolution: typing.ClassVar[HighsCallbackType]
    kCallbackSimplexInterrupt: typing.ClassVar[HighsCallbackType]
    kNumCallbackType: typing.ClassVar[HighsCallbackType]
    def __eq__(self, other: typing.Any) -> bool: ...
    def __getstate__(self) -> int: ...
    def __hash__(self) -> int: ...
    def __index__(self) -> int: ...
    def __init__(self, value: int) -> None: ...
    def __int__(self) -> int: ...
    def __ne__(self, other: typing.Any) -> bool: ...
    def __repr__(self) -> str: ...
    def __setstate__(self, state: int) -> None: ...
    def __str__(self) -> str: ...
    @property
    def name(self) -> str: ...
    @property
    def value(self) -> int: ...

kCallbackIpmInterrupt: HighsCallbackType
kCallbackLogging: HighsCallbackType
kCallbackMax: HighsCallbackType
kCallbackMin: HighsCallbackType
kCallbackMipDefineLazyConstraints: HighsCallbackType
kCallbackMipGetCutPool: HighsCallbackType
kCallbackMipImprovingSolution: HighsCallbackType
kCallbackMipInterrupt: HighsCallbackType
kCallbackMipLogging: HighsCallbackType
kCallbackMipSolution: HighsCallbackType
kCallbackSimplexInterrupt: HighsCallbackType
kNumCallbackType: HighsCallbackType



# An example branch-and-price algorithm for the bin packing problem
#
from collections import defaultdict
from operator import itemgetter
import highspy, random, time, math

# Relative gap for column generation
CG_GAP_REL = 1e-4

# Random seeds for reproducibility
random.seed(100)
SEED = 100

# Instance parameters
NumberItems = 40
ItemWeights = [round(random.uniform(1, 10), 1) for _ in range(NumberItems)]
BinCapacity = 15


#
# Solve instance with greedy first fit decreasing (FFD) heuristic
#
def solveGreedyModel():
    bins = defaultdict(float)
    solution = defaultdict(list)

    for item, w in sorted(enumerate(ItemWeights), reverse=True, key=itemgetter(1)):
        index = next((i for i, W in bins.items() if W + w <= BinCapacity), len(bins))
        bins[index] += w
        solution[index].append(item)
    
    return list(solution.values())

#
# The compact bin-packing model (for comparison with branch-and-price)
#
# min sum_{j} y_{j}
# s.t. 
#   sum_{j} x_{ij} == 1, \forall i                 // Each item i is assigned to exactly one bin
#   sum_{i} w_{i} * x_{ij} <= c * y_{j}, \forall j // Bin capacity constraint
#
#   x_{ij} = 1 if item i is assigned to bin j, 0 otherwise
#   y_{j}  = 1 if bin j is open, 0 otherwise
#
def solveCompactModel():
    # Estimate maximum number of bins with first fit decreasing (for smaller sized model)
    greedy_bins = solveGreedyModel()
    B = len(greedy_bins)

    m = highspy.Highs()
    m.setOptionValue('output_flag', True)
    m.setOptionValue('mip_abs_gap', 1-1e-5) # Since all variables in the objective are binary, it should stop when the absolute gap is below 1
    m.setOptionValue('random_seed', SEED)

    # add binary variables x_{ij} and y_{j}
    x = {(i,j): i*B + j for i in range(NumberItems) for j in range(B) }
    y = [len(x) + j for j in range(B)]

    m.addVars(len(x), [0]*len(x), [1]*len(x));  # x_{ij} \in {0,1}, \forall i,j
    m.addVars(len(y), [0]*len(y), [1]*len(y));  # y_{j}  \in {0,1}, \forall j
    m.changeColsIntegrality(
        len(x)+len(y),
        list(range(len(x)+len(y))),
        [highspy.HighsVarType.kInteger]*(len(x)+len(y))
    )

    # min sum_{j} y_{j}
    m.changeObjectiveSense(highspy.ObjSense.kMinimize)
    m.changeColsCost(len(y), y, [1]*len(y))
    
    #      \sum_{j} x_{ij} == 1, \foreach i 
    # 1 <= \sum_{j} x_{ij} <= 1
    for i in range(NumberItems):
        m.addRow(
            1,                                              # lhs
            1,                                              # rhs
            B,                                              # Number of non-zero variables
            [x[i,j] for j in range(B)],                     # Indexes of variable
            [1] * B                                         # Coefficients of variables
        )

    #         sum_{i} w_{i} * x_{ij} <= c * y_{j}, \foreach j
    # -inf <= sum_{i} w_{i} * x_{ij}  - c * y_{j} <= 0
    for j in range(B):
        m.addRow(
            -highspy.kHighsInf,                                             # lhs
            0,                                                              # rhs
            NumberItems+1,                                                  # Number of non-zero variables 
            [x[i,j] for i in range(NumberItems)] + [y[j]],                  # Indexes of variable
            [ItemWeights[i] for i in range(NumberItems)] + [-BinCapacity]   # Coefficients of variables  
        )
        
    m.run()
    vals = list(m.getSolution().col_value)

    bins = [
        [i for i in range(NumberItems) if vals[x[i,j]] > 0.9]
        for j in range(B)
        if vals[y[j]] > 0.9
    ]
    
    return bins

#
# Master problem for column generation
# A column represents a set of items packed into a bin
#
# min \sum_{k} \lambda_{k}
# s.t. 
#   \sum_{k \in K_i} \lambda_{k} == 1, \foreach i  (\mu_i duals)  // Each item i is packed exactly once (but may be spread over multiple columns)
#                                                                 //   where K_i is the set of columns where i appears
#   0 <= \lambda_{k} <= 1, \foreach k                             // Can use each column at most once (can be fractional)
#
def createMasterProblem(columns: list):
    m = highspy.Highs()
    m.setOptionValue('output_flag', False)
    m.setOptionValue('random_seed', SEED)
    
    m.addVars(len(columns), [0]*len(columns), [1]*len(columns))
    
    # min \sum_{k} \lambda_{k}
    m.changeObjectiveSense(highspy.ObjSense.kMinimize)
    m.changeColsCost(len(columns), list(range(len(columns))), [1]*len(columns))
    
    #      \sum_{k \in K_i} \lambda_{k} == 1, \foreach i
    # 1 <= \sum_{k \in K_i} \lambda_{k} <= 1
    for i in range(NumberItems):
        K_i = [k for k, column in enumerate(columns) if i in column]
        m.addRow(1, 1, len(K_i), K_i, [1]*len(K_i))

    m.run() 

    solution = m.getSolution()
    vals = list(solution.col_value)
    duals = list(solution.row_dual)

    return m, vals, duals

#
# Solve the knapsack subproblem exactly with HiGHS 
#
# 1 - max \sum_{i} \mu_{i} * z_{i}
# s.t.
#   \sum_{i} w_{i} * z_{i} <= capacity
#
#   z_{i} = 1 if item i is packed, 0 otherwise
#
def solveSubproblemExact(duals):
    m = highspy.Highs()
    m.setOptionValue('output_flag', False)
    m.setOptionValue('random_seed', SEED)

    m.addVars(NumberItems, [0]*NumberItems, [1]*NumberItems)
    m.changeColsIntegrality(NumberItems, list(range(NumberItems)), [highspy.HighsVarType.kInteger]*NumberItems)

    # max \sum_{i} \mu_{i} * z_{i}
    # where \mu_{i} is the dual variable for the i-th row of the master problem
    m.changeColsCost(NumberItems, list(range(NumberItems)), duals)
    m.changeObjectiveSense(highspy.ObjSense.kMaximize)

    #         sum_{i} w_{i} * z_{i} <= capacity, \foreach j
    # -inf <= sum_{i} w_{i} * z_{i} <= capacity
    m.addRow(
        -highspy.kHighsInf,
        BinCapacity,
        NumberItems,
        list(range(NumberItems)),
        ItemWeights
    )
    
    m.run()

    vals = list(m.getSolution().col_value)
    new_column = sorted([i for i in range(NumberItems) if vals[i] > 0.9])

    return 1 - m.getObjectiveValue(), new_column


# Solve the knapsack subproblem with greedy heuristic
def solveSubproblemNotExact(duals):
    total_weight = 0
    new_column = []
    
    for i in sorted(range(NumberItems), key=lambda i: -duals[i]/ItemWeights[i]):
        if duals[i] >= 0 and ItemWeights[i] + total_weight <= BinCapacity:
            total_weight += ItemWeights[i]
            new_column += [i]
    
    return 1 - sum(duals[i] for i in new_column), sorted(new_column)


#
# Generate columns for the master problem
#
def generateColumns(columns: list, m, msg=True, solve_exact = False, start_time = 0):
    best_gap = math.inf

    iter = 0
    while True:
        solution = m.getSolution()
        duals = list(solution.row_dual)
        ub = m.getObjectiveValue()

        # solve sub problem to generate new column
        if not solve_exact:
            obj_sub, new_column = solveSubproblemNotExact(duals)
        else:
            obj_sub, new_column = solveSubproblemExact(duals)

        iter += 1     
        obj = ub + min(0, obj_sub)  # we terminate if obj_sub >= 0
        gap = (ub-obj)/ub
        
        if gap < best_gap:
            best_gap = gap
            if msg:
                row = [
                    f"{len(columns)}",
                    f"{round(obj_sub, 3) :.4f}",
                    f'{round(obj, 3) :.4f}',
                    f"{round(ub, 3) :.4f}",
                    f"{max(0, gap) :.3%}", 
                    f"{round(time.perf_counter()-start_time, 2) :.2f}"
                ]
                if iter == 1:
                    header = ["Columns", 'Pricing', 'Obj', 'UB', 'gap', 'Time']
                    row_format = "".join([
                        "{:>"+str(max(len(row_el), len(header_el))+3)+"}"
                        for row_el, header_el in zip(row, header)
                    ])
                    print(row_format.format(*header))

                print(row_format.format(*row))

        # terminate if no column with good reduced cost is found
        if obj_sub >= 0 or gap < CG_GAP_REL:
            break

        # Adds a new column to the master problem
        columns += [new_column]
        m.addCol(
            1,                      # cost
            0,                      # lower bound
            1,                      # upper bound
            len(new_column),        # number of rows
            new_column,             # indexes of rows
            [1]*len(new_column)     # coefficients of rows
        )

        # Solves the master problem
        m.run()

    return m, columns, list(solution.col_value)


#
# Branching helper structures/functions
#
class Node:
    layer: int
    assigned_columns: list[int]

    value: float
    final_columns: list[int]
    final_columns_vals: list[float]
    final_fractional_columns: list[int]

    def __init__(self, layer: int, assigned_columns: list[int], parent = None):
        self.layer = layer
        self.parent = parent
        self.assigned_columns = assigned_columns


def getFractionalColumns(vals: list, columns: list):
    return sorted(
        [k for k, val in enumerate(vals) if abs(round(val) - val) > 1e-6],
        key=lambda k: -len(columns[k]) # order by most fractional (largest number of items)
    )


def solveNode(node: Node, columns: list[list[int]], m):
    # "enforce" the branch constraints, i.e., selected columns in the current node
    lower_bounds = [int(k in node.assigned_columns) for k in range(len(columns))]

    m.changeColsBounds(
        len(columns),                       # Qty columns to change bounds
        list(range(len(columns))),          # Which columns
        lower_bounds,                       # Lower bound (0 if not assigned else 1)
        [1]*len(columns)                    # Upper bound (always 1)
    )

    # Making sure that there is no repeated index
    assert len(node.assigned_columns) == len(set(node.assigned_columns))

    m.run()

    if m.getModelStatus() == highspy.HighsModelStatus.kOptimal:        
        m, columns, vals = generateColumns(columns, m, msg=False, solve_exact=False) # Generate columns quickly
        m, columns, vals = generateColumns(columns, m, msg=False, solve_exact=True)  # Prove optimality on node

        node.value = sum(vals)
        node.final_columns = columns
        node.final_columns_vals = vals
        node.final_fractional_columns = getFractionalColumns(vals, columns)

        # ensure no column already assigned to 1 in the node is returned back with a fractional value
        assert len(set(node.final_fractional_columns).intersection(set(node.assigned_columns))) == 0
        return True, m, node
    
    return False, m, node


#
# Branch-and-price algorithm
#
# Note: Branch selection has been chosen for simple implementation and it's ability to avoid symmetry. 
# This comes at the cost of an unbalanced search tree (other approaches may solve the problem with fewer branches), 
# and it is less likely to be generalizable to other problems.
#
# Specifically, the code "up-branches" columns with fractional value in the RMP solution, i.e., forces specific columns 
# to be selected in RMP.  The subproblems don't need to explicitly enforce this constraint (unlike other branching strategies).
#
def branchAndPrice(m, vals, columns, start_time=0):    
    header = ["NodesExpl", "TreeSize", "CurrCols", "FracCols", "UB", "Time"]
    row_format ="{:>12}" * (len(header))
    print(row_format.format(*header))

    root_node = Node(-1, [])
    root_node.value = sum(vals)
    root_node.final_columns = columns
    root_node.final_columns_vals = vals
    root_node.final_fractional_columns = getFractionalColumns(vals, columns)

    # create initial branches for each column with fractional value in RMP solution
    branch_tree: list[Node] = [Node(0, [column_idx]) for column_idx in root_node.final_fractional_columns]
    branch_tree_dict = {frozenset(node.assigned_columns): node for node in branch_tree}

    # RMP lower bound can be rounded up to nearest integer (avoiding floating point precision issues)
    # so, take integer floor, add one if fractional part greater than tolerance
    rmp_LB = int(root_node.value) + int(root_node.value % 1 > 1e-6)
    best_obj = math.inf
    best_node = None
    best_count_frac_columns = math.inf
    count_nodes_visited = 0

    # if no fractional columns, then root node is an optimal integer solution
    if len(root_node.final_fractional_columns) == 0:
        best_node = root_node
        print(row_format.format(*[0, 0, len(columns), 0, best_node.value, f"{round(time.perf_counter()-start_time, 2) :.2f}" ]))

    # explore branches, looking for optimal integer solution
    while len(branch_tree) > 0:
        # Choose next node in branch to evaluate.  Prioritize nodes that are likely to be 
        # integer (DFS with fewer fractional columns), and likely to be optimal (larger value).
        #
        # i.e., maximize branch depth (layer), minimize number of fractional columns of its parent, 
        # and maximize value of its original fractional columns.
        node = min(branch_tree, key=lambda node: (
            -node.layer,
            len(node.parent.final_fractional_columns) if node.parent is not None else math.inf,
            -node.parent.final_columns_vals[node.assigned_columns[-1]] if node.parent is not None else 0
        ))
        branch_tree.remove(node)

        # Solves the node with column generation
        mp_is_feasible, m, node = solveNode(node, columns, m)
        count_nodes_visited += 1

        # Only add columns if the master problem is feasible
        # Prune nodes that cannot lead to optimal solution        
        if mp_is_feasible and node.value < best_obj:
            columns = node.final_columns            
            count_fractional_columns = len(node.final_fractional_columns)

            # if no fractional columns, then this is an integer solution
            # update the best solution here for console output
            if count_fractional_columns == 0 and node.value < best_obj:
                best_obj = int(round(node.value, 0))  # avoid float precision issues
                best_node = node

            # found a less fractional solution or enough time has elapsed
            if count_fractional_columns < best_count_frac_columns or time.perf_counter()-start_time > 5:
                best_count_frac_columns = count_fractional_columns
                print(row_format.format(*[
                    count_nodes_visited,
                    len(branch_tree),
                    len(columns),
                    best_count_frac_columns,
                    best_obj if best_obj < math.inf else "-",
                    f"{round(time.perf_counter()-start_time, 2) :.2f}"
                ]))

            # add all fractional columns as new branches
            if count_fractional_columns > 0:
                for column_idx in node.final_fractional_columns:
                    new_assigned_columns = node.assigned_columns + [column_idx]
                    key = frozenset(new_assigned_columns)
                    
                    if key not in branch_tree_dict:
                        new_node = Node(node.layer + 1, new_assigned_columns, node)
                        branch_tree = [new_node] + branch_tree
                        branch_tree_dict[key] = new_node

            # if no fractional columns, then this is an integer solution
            # stop if we've found a provably optimal solution (using lower bound from RMP root node)
            elif rmp_LB == best_obj:
                break

    # explored the entire tree, so best found solution is optimal
    if best_node is not None:
        return [column for idx, column in enumerate(best_node.final_columns) if best_node.final_columns_vals[idx] > 0.9]
    
    raise Exception("It should not get to this point")


if __name__ == '__main__':
    if max(ItemWeights) > BinCapacity:
        print(f"Instance is infeasible: item {max(enumerate(ItemWeights), key=itemgetter(1))[0]} has weight {max(ItemWeights)} > {BinCapacity} (bin capacity).")
        exit()

    start_time = time.perf_counter()
    greedy_bins = solveGreedyModel()
    greedy_time = time.perf_counter()-start_time
    print(f"Greedy estimate: {len(greedy_bins)} bins")
    print(f"Finished in {greedy_time: .2f}s\n")

    start_time = time.perf_counter()
    compact_bins = solveCompactModel()
    compact_time = time.perf_counter()-start_time

    print(f"\nSolution by compact model: {len(compact_bins)} bins")
    for bin, items in enumerate(compact_bins):
        tt_weight = round(sum(ItemWeights[i] for i in items))
        print(f"Bin {bin+1} ({tt_weight} <= {BinCapacity}): {items}")
        assert tt_weight <= BinCapacity

    print(f"Finished in {compact_time: .2f}s\n")

    # start with initial set of columns for feasible master problem
    start_time = time.perf_counter()
    columns = [[i] for i in range(NumberItems)] + greedy_bins
    m, vals, duals = createMasterProblem(columns)
    
    print("\nSolving root node:")
    m, columns, vals = generateColumns(columns, m, solve_exact=False, start_time=start_time)
    
    print("\nProving optimality on root node:")
    m, columns, vals = generateColumns(columns, m, solve_exact=True, start_time=start_time)

    print("\nBranch-and-price:")
    cg_bins = branchAndPrice(m, vals, columns, start_time)
    cg_time = time.perf_counter()-start_time

    print(f"\nSolution by column generation: {len(cg_bins)} bins")
    for bin, items in enumerate(cg_bins):
        tt_weight = round(sum(ItemWeights[i] for i in items))
        print(f"Bin {bin+1} ({tt_weight} <= {BinCapacity}): {items}")
        assert tt_weight <= BinCapacity

    print(f"Finished in {cg_time: .2f}s\n")

    print(f"Greedy : {len(greedy_bins)} bins, {greedy_time:6.2f}s")
    print(f"Compact: {len(compact_bins)} bins, {compact_time:6.2f}s")
    print(f"ColGen : {len(cg_bins)} bins, {cg_time:6.2f}s\n")

